from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest
import structlog
from cloudevents import conversion
from flask import Flask

from ucam_faas import FaaSGunicornApplication, cloud_event, raw_event
from ucam_faas.exceptions import UCAMFAASCouldNotProcess
from ucam_faas.testing import CloudEventFactory, CreateEventAppClientFn

if TYPE_CHECKING:
    from typing_extensions import Never


# Raw events
@raw_event
def example_raw_event_no_exception(event: bytes) -> None:
    pass


@raw_event
def example_raw_event_handled_exception(event: bytes) -> Never:
    raise UCAMFAASCouldNotProcess


@raw_event
def example_raw_event_unhandled_exception(event: bytes) -> Never:
    raise Exception("Did not expect this")


# Cloud Events
@cloud_event
def example_cloud_event_no_exception(event: Any) -> None:
    pass


@cloud_event
def example_cloud_event_handled_exception(event: Any) -> None:
    raise UCAMFAASCouldNotProcess


@cloud_event
def example_cloud_event_unhandled_exception(event: Any) -> None:
    raise Exception("Did not expect this")


def test_faas_gunicorn_application_bind() -> None:
    app = Flask(__name__)
    application = FaaSGunicornApplication(app, "0.0.0.0", "8080")

    with patch("gunicorn.app.base.BaseApplication.run") as mock_run:
        application.run()
        mock_run.assert_called_once()  # Ensures that the server's run method was indeed called


@pytest.mark.parametrize(
    "target_tuple",
    [
        (
            "example_raw_event_no_exception",
            "example_raw_event_handled_exception",
            "example_raw_event_unhandled_exception",
        ),
        (
            "example_cloud_event_no_exception",
            "example_cloud_event_handled_exception",
            "example_cloud_event_unhandled_exception",
        ),
    ],
)
def test_exceptions_raw_events(
    event_app_test_client_factory: CreateEventAppClientFn, target_tuple: tuple[str, str, str]
) -> None:
    this_file_path = "ucam_faas/tests/test_ucam_faas.py"

    # Both raw and cloud event functions except cloud eventd
    valid_cloud_event = conversion.to_dict(CloudEventFactory.build())

    # No exception
    test_client = event_app_test_client_factory(target=target_tuple[0], source=this_file_path)
    response = test_client.post("/", json=valid_cloud_event)
    assert response.status_code == 200

    # Handle exception
    test_client = event_app_test_client_factory(target=target_tuple[1], source=this_file_path)
    response = test_client.post("/", json=valid_cloud_event)
    assert response.status_code == 500
    assert "The function raised UCAMFAASCouldNotProcess" in response.data.decode()

    # Unhandled exception
    test_client = event_app_test_client_factory(target=target_tuple[2], source=this_file_path)
    with structlog.testing.capture_logs() as cap_logs:
        response = test_client.post("/", json=valid_cloud_event)
    assert response.status_code == 500

    assert len(cap_logs) == 1
    log_call = cap_logs[0]
    assert (
        log_call.get("event") == "function_failed_uncaught_exception"
        # structlog will include exception details in rendered log message
        and log_call.get("exc_info") is True
    )



from __future__ import annotations

import os.path
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast

import click
import flask
import flask.typing
import functions_framework
import gunicorn.app.base  # type: ignore[import-untyped]
from cloudevents.http.event import CloudEvent
from structlog.typing import FilteringBoundLogger
from typing_extensions import ParamSpec
from ucam_observe import get_structlog_logger  # type: ignore[import-untyped]
from ucam_observe.gunicorn import logconfig_dict  # type: ignore[import-untyped]
from werkzeug.exceptions import InternalServerError

from ucam_faas.exceptions import UCAMFAASException

if TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication

P = ParamSpec("P")
T = TypeVar("T")

# As well as making a logger available this should setup logging before the flask app is created
logger: FilteringBoundLogger = get_structlog_logger(__name__)


def _common_function_wrapper(function: Callable[P, T]) -> Callable[P, T]:
    def _common_function_wrapper_internal(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return function(*args, **kwargs)
        except UCAMFAASException as exception:
            exception_name = exception.__class__.__name__

            logger.warning("function_failed_gracefully", exception_name=exception_name)

            raise InternalServerError(description=f"The function raised {exception_name}.")

        except Exception as exception:
            exception_name = exception.__class__.__name__

            logger.exception("function_failed_uncaught_exception", exception_name=exception_name)

            raise exception

    return _common_function_wrapper_internal


class RawEventHandlerFn(Protocol):
    def __call__(self, event: bytes, /) -> flask.typing.ResponseReturnValue | None: ...

    @property
    def __name__(self) -> str: ...


class RegisteredRawEventHandlerFn(Protocol):
    __wrapped__: RawEventHandlerFn

    def __call__(self, request: flask.Request, /) -> flask.typing.ResponseReturnValue: ...

    @property
    def __name__(self) -> str: ...


def raw_event(function: RawEventHandlerFn) -> RegisteredRawEventHandlerFn:
    @_common_function_wrapper
    def _raw_event_internal(request: flask.Request, /) -> flask.typing.ResponseReturnValue:
        return_value = function(request.data)

        if return_value is not None:
            return return_value

        return "", 200

    # Decorators must preserve the wrapped function identity because
    # functions_framework registers metadata against the __name__ of `function`.
    _raw_event_internal.__name__ = function.__name__
    _raw_event_internal = functions_framework.http(_raw_event_internal)

    _raw_event_internal = cast(RegisteredRawEventHandlerFn, _raw_event_internal)
    _raw_event_internal.__wrapped__ = function

    return _raw_event_internal


class CloudEventHandlerFn(Protocol):
    def __call__(self, event_data: Any, /) -> None: ...

    @property
    def __name__(self) -> str: ...


class RegisteredCloudEventHandlerFn(Protocol):
    __wrapped__: CloudEventHandlerFn

    def __call__(self, event: CloudEvent, /) -> None: ...

    @property
    def __name__(self) -> str: ...


def cloud_event(function: CloudEventHandlerFn) -> RegisteredCloudEventHandlerFn:
    @_common_function_wrapper
    def _cloud_event_internal(event: CloudEvent, /) -> None:
        return function(event.data)

    # Decorators must preserve the wrapped function identity because
    # functions_framework registers metadata against the __name__ of `function`.
    _cloud_event_internal.__name__ = function.__name__
    _cloud_event_internal = functions_framework.cloud_event(_cloud_event_internal)

    _cloud_event_internal = cast(RegisteredCloudEventHandlerFn, _cloud_event_internal)
    _cloud_event_internal.__wrapped__ = function

    return _cloud_event_internal


class FaaSGunicornApplication(gunicorn.app.base.Application):  # type: ignore[misc] # gunicorn is not typed # noqa: E501
    def __init__(self, app: WSGIApplication, host: str, port: int | str) -> None:
        self.host = host
        self.port = port
        self.app = app

        self.options = {
            "bind": "%s:%s" % (host, port),
            "workers": os.environ.get("WORKERS", 2),
            "threads": os.environ.get("THREADS", (os.cpu_count() or 1) * 4),
            "timeout": 0,
            "limit_request_line": 0,
            "logconfig_dict": logconfig_dict,
        }

        super().__init__()

    def load_config(self) -> None:
        for key, value in self.options.items():
            self.cfg.set(key, value)

    def load(self) -> WSGIApplication:
        return self.app


def _initialize_ucam_faas_app(target: str, source: str | Path | None, debug: bool) -> flask.Flask:
    app: flask.Flask = functions_framework.create_app(target, source)  # type: ignore[no-untyped-call] # noqa: E501

    app.logger.info("flask_app_created")

    @app.route("/healthy")
    @app.route("/status")
    def get_status() -> str:
        return "ok"

    return app


def run_ucam_faas(
    target: str, source: str | Path | None, host: str, port: int, debug: bool
) -> None:  # pragma: no cover
    app = _initialize_ucam_faas_app(target, source, debug)
    if debug:
        app.run(host, port, debug)
    else:
        server = FaaSGunicornApplication(app, host, port)
        server.run()


@click.command()
@click.option("--target", envvar="FUNCTION_TARGET", type=click.STRING, required=True)
@click.option("--source", envvar="FUNCTION_SOURCE", type=click.Path(), default=None)
@click.option("--host", envvar="HOST", type=click.STRING, default="0.0.0.0")
@click.option("--port", envvar="PORT", type=click.INT, default=8080)
@click.option("--debug", envvar="DEBUG", is_flag=True)
def _cli(target: str, source: str, host: str, port: int, debug: bool) -> None:  # pragma: no cover
    run_ucam_faas(target, source, host, port, debug)



from cloudevents.pydantic.v2 import CloudEvent

from ucam_faas.testing import CloudEventFactory


def test_cloudeventfactory_model_static_type_has_pydantic_methods() -> None:
    event = CloudEventFactory.build()
    assert CloudEvent.model_validate(event.model_dump()) == event

    # However type checkers don't see Pydantic methods on cloudevents.pydantic.CloudEvent
    from cloudevents.pydantic import CloudEvent as CloudEventAutoVersion

    # mypy complains: "type[CloudEvent]" has no attribute "model_validate"
    assert CloudEventAutoVersion.model_validate(event.model_dump()) == event  # type: ignore[attr-defined] # noqa: E501



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Protocol

from flask.testing import FlaskClient

from . import _initialize_ucam_faas_app

if TYPE_CHECKING:
    from typing_extensions import Never


class CreateEventAppClientFn(Protocol):
    def __call__(self, target: str, source: str | Path | None = None) -> FlaskClient:
        ...


try:
    if TYPE_CHECKING:
        from cloudevents.pydantic.v2 import CloudEvent
    else:
        from cloudevents.pydantic import CloudEvent
    from polyfactory.factories.pydantic_factory import ModelFactory
    from pytest import fixture

    @fixture
    def event_app_test_client_factory() -> CreateEventAppClientFn:
        def _event_app_client(target: str, source: str | Path | None = None) -> FlaskClient:
            test_app = _initialize_ucam_faas_app(target, source, True)
            return test_app.test_client()

        return _event_app_client

    class CloudEventFactory(ModelFactory[CloudEvent]):
        __model__ = CloudEvent

        specversion = "1.0"

except ImportError as e:
    _import_error = e

    @fixture
    def event_app_test_client_factory() -> Never:
        raise RuntimeError(
            f"Fixture {event_app_test_client_factory.__name__} is not available "
            f"because ucam_faas is not installed with the 'testing' extra]"
        ) from _import_error
